#include "TimeProjection.hh"
#include "WorldBaseReport.hh"

using namespace std;
using namespace boost::numeric;
using namespace boost::numeric::interval_lib;

bool TimeProjection::canReachTaskFromTo(TaskListNode* t1, TaskListNode* t2) {
	if(t1->isAPredecessorOf(t2)) return true;
	else {
		vector<TaskListNode*> nextTasks = t1->getSuccessors();
		for(unsigned long int i = 0 ; i < nextTasks.size() ; i++) {
			if(canReachTaskFromTo(nextTasks[i], t2)) return true;
		}
		return false;
	}
}

bool TimeProjection::isCompatible(unsigned long int i1, unsigned long int i2) {
	vector<string> dataP, dataE;
	vector<string>::iterator it1, it2;
	if(i1 == i2) return true;
	for(it2=effectsReports[i2].begin() ; it2 != effectsReports[i2].end() ; it2++) {
		dataE = parseModificationReport(*it2);
		for(it1=precondsReports[i1].begin() ; it1 != precondsReports[i1].end() ; it1++) {
			dataP = parseTestReport(*it1);
			if(dataE[1]==dataP[1] && dataE[2]==dataP[2]){ // same entity and attribute
				if(dataE[0]==ATOM_CHANGE) { // for atom
					if(dataE[4]!=dataP[3] && dataP[0] == ATOM_TEST_EQUAL) return false;
					if(dataE[4]==dataP[3] && dataP[0] == ATOM_TEST_DIFF) return false;
				}
				else {
					if(dataE[3]==dataP[3]) { // same value (for set testing)
						if(dataE[0] == VECTOR_REMOVE && dataP[0] == VECTOR_TEST_IN) return false;
						if(dataE[0] == VECTOR_ADD && dataP[0] == VECTOR_TEST_NOT_IN) return false;
					}
				}
			}
		}
	}
	return true;
}

void TimeProjection::updateLinksCurrentTask(unsigned long int projID) {
	// peut-etre retenir liste objets intervenant dans liens pour supervision
	vector<string> dataP, dataE;
	vector<string>::iterator it1, it2;
	for(long int j = projID - 1 ; j >= 0 ; j--) {
		bool detectedLink = false;
		for(it2=effectsReports[j].begin() ; it2 != effectsReports[j].end() ; it2++) {
			dataE = parseModificationReport(*it2);
			for(it1=precondsReports[projID].begin() ; it1 != precondsReports[projID].end() ; it1++) {
				dataP = parseTestReport(*it1);
				if(dataE[1]==dataP[1] && dataE[2]==dataP[2]){ // same entity and attribute
					if(dataE[0]==ATOM_CHANGE) { // atom change
						if(dataE[3]==dataP[3] && dataP[0] == ATOM_TEST_DIFF) { // testing old value
							detectedLink = true;
							break;
						}
						if(dataE[4]==dataP[3] && dataP[0] == ATOM_TEST_EQUAL) { // testing new value
							detectedLink = true;
							break;
						}
					}
					else {
						if(dataE[3]==dataP[3]) { // same value (for set testing)
							if(dataE[0] == VECTOR_ADD && dataP[0] == VECTOR_TEST_IN) { // value in
								detectedLink = true;
								break;
							}
							if(dataE[0] == VECTOR_REMOVE && dataP[0] == VECTOR_TEST_NOT_IN) { // val not in
								detectedLink = true;
								break;
							}
						}
					}
				}
			}
			if(detectedLink) {
				break;
			}
	}
		if(detectedLink) {
			TaskListNode* t1 = projection.find(projectionOrder[j]);
			TaskListNode* t2 = projection.find(projectionOrder[projID]);
//			for(unsigned long int k = j+1 ; k < projID ; k++) {
//				if(isCompatible(k,projID)) {
//					t1 = projection.find(projectionOrder[k-1]);
//					break;
//				}
//			}	
			if(!this->canReachTaskFromTo(t1, t2)) {
				t1->AddSuccessor(t2, LOW_LEVEL_LINK);
			}
		}
	}
}

void TimeProjection::updateLinks(LightWorldBase& wb_at_start) {
	LightWorldBase wb_copy(wb_at_start); 
	TaskList TL_copy(projection); 
	TL_copy.setAllStatus(TODO);
	vector<string> precondsReport, data;
	vector<string>::iterator it;
	WorldBaseReport report;
	for(unsigned int i=0 ; i<projectionOrder.size() ; i++) {
		report.clearAll();
		TL_copy.find(projectionOrder[i])->getInstance().checkPreconditionsWithReport(wb_copy, report);
		precondsReport = report.getTestsReport();
		for(it=precondsReport.begin() ; it != precondsReport.end() ;) {
			data = parseTestReport(*it);
			if(!(wb_copy.find(data[1])->isDynamic(data[2]))) it = precondsReport.erase(it);
			else it++;
		}
		precondsReports[i] = precondsReport;
		report.clearAll();
		TL_copy.project(projectionOrder[i]);
		TL_copy.find(projectionOrder[i])->getInstance().applyActionEffectsWithReport(wb_copy, report);
		effectsReports[i] = report.getModificationsReport();
	}
	for(unsigned int i=0 ; i<projectionOrder.size() ; i++) {
		this->updateLinksCurrentTask(i);
	}	
}

pair<double, double> TimeProjection::findMaxTime(vector<TaskListNode*> tln) {
	pair<double, double> maxTime = make_pair(.0, .0);
	for(unsigned int i = 0; i < tln.size() ; i++) {
		pair<double, double> tlnTime = projectionTimes[tln[i]->getID()];
		if(maxTime.second < tlnTime.second) maxTime = tlnTime;
	}
	return maxTime;
}

pair<double, double> TimeProjection::getTimeWindow(unsigned long int id) const {
	return projectionTimes.find(id)->second;
}

void TimeProjection::computeTimeWindows(){
	for(unsigned int i = 0 ; i < projectionOrder.size() ; i++) {
		TaskListNode* currentNode = projection.find(projectionOrder[i]);
		if(currentNode->getPredecessors().size() == 0) {
			projectionTimes[projectionOrder[i]].first = .0;
			projectionTimes[projectionOrder[i]].second = currentNode->getInstance().getMaxTime();
			taskAtTimeMin[0].push_back(projectionOrder[i]);
		}
		else {
			pair<double, double> currentTime = this->findMaxTime(currentNode->getPredecessors());
			projectionTimes[projectionOrder[i]].first = currentTime.second;
			projectionTimes[projectionOrder[i]].second = currentTime.second + currentNode->getInstance().getMaxTime();
			taskAtTimeMin[currentTime.first].push_back(projectionOrder[i]);
		}
	}	
}

pair<double, double> TimeProjection::totalDurationWindow() const {
	double firstTime = .0, maxTime = .0;
	if(projectionOrder.size() > 0) firstTime = projectionTimes.find(projectionOrder[0])->second.first;
	for(unsigned int i = 0 ; i < projectionOrder.size() ; i ++) {
		if(projectionTimes.find(projectionOrder[i])->second.second > maxTime) {
			maxTime = projectionTimes.find(projectionOrder[i])->second.second;
		}
	}
	return make_pair(firstTime, maxTime);
}

double TimeProjection::totalDuration() const {
	return this->totalDurationWindow().second;
}

vector<double> TimeProjection::inactiveDuration(vector<unsigned long int> const& actionIDs, double& fromStart, double& toEnd) const {
	double duration = .0;
	vector<double> inactiveDurations;
	pair<double, double> currentT, nextT;
	if(actionIDs.empty()) { fromStart = 0.; toEnd = 0.;}
	else {
		for(unsigned int i=0 ; i<actionIDs.size()-1 ; ++i) {
			duration = 0;
			currentT = projectionTimes.find(actionIDs[i])->second; 
			nextT = projectionTimes.find(actionIDs[i+1])->second;
			duration += nextT.first - currentT.second;
			if(duration > 0) inactiveDurations.push_back(duration); 
		}
		// from start and to end
		pair<double, double > tdw = this->totalDurationWindow();
		currentT = projectionTimes.find(actionIDs[0])->second;
		fromStart = currentT.first - tdw.first;
		currentT = projectionTimes.find(actionIDs[actionIDs.size()-1])->second;
		toEnd = tdw.second - currentT.second;
	}
	return inactiveDurations;
}

bool TimeProjection::possibleState(WorldBaseReport const& WBR, LightWorldBase const& wb_at_start, vector<double>& durations, bool& atEnd) const {
	vector<TInterval> stateResult;
	vector<vector<string> > stateData; 
	vector<string> stateReport = WBR.getTestsReport();
	for(unsigned int i = 0 ; i < stateReport.size() ; i++) {
		stateData.push_back(parseTestReport(stateReport[i]));
	}
	LightWorldBase wb_current(wb_at_start);
	TaskList tl_current(projection); 
	tl_current.setAllStatus(TODO);
	WorldBaseReport report;
	bool containState;
	map<double, vector<unsigned long int> >::const_iterator it;
	for(it = taskAtTimeMin.begin() ; it != taskAtTimeMin.end(); it++) {
		for(unsigned int j = 0 ; j < it->second.size() ; j++) {
			containState = true;
			for(unsigned int k = 0 ; k < stateReport.size() ; k++) {
				vector<string> vecData = stateData[k];
				set<string> mySet = (*(wb_current.find(vecData[1])))(vecData[2]);
				if((vecData[0] == ATOM_TEST_EQUAL && (*(wb_current.find(vecData[1])))[vecData[2]] == vecData[3]) ||
					(vecData[0] == ATOM_TEST_DIFF && (*(wb_current.find(vecData[1])))[vecData[2]] != vecData[3])) ;
				else if((vecData[0] == VECTOR_TEST_IN && find(mySet.begin(), mySet.end(), vecData[3] ) != mySet.end()) ||
					(vecData[0] == VECTOR_TEST_NOT_IN && find(mySet.begin(), mySet.end(), vecData[3] ) == mySet.end())) ;
				else containState = false;
			}
			tl_current.project(it->second[j]); // effects hold at the end of the action
			tl_current.find(it->second[j])->getInstance().applyActionEffectsWithReport(wb_current, report);
			if(containState) {
				pair<double, double> currentT = projectionTimes.find(it->second[j])->second;
				pair<double, double> durationT = make_pair(projection.find(it->second[j])->getInstance().getMinTime(), 
					projection.find(it->second[j])->getInstance().getMaxTime());
				double endT = (currentT.first + currentT.second) / 2;
				double debT = endT - ((durationT.first + durationT.second) / 2);
				stateResult.push_back(TInterval(debT, endT));
			}
		}
	}
	containState = true; // test for last state (atEnd)
	for(unsigned int k = 0 ; k < stateReport.size() ; k++) {
		vector<string> vecData = stateData[k];
		set<string> mySet = (*(wb_current.find(vecData[1])))(vecData[2]);
		if((vecData[0] == ATOM_TEST_EQUAL && (*(wb_current.find(vecData[1])))[vecData[2]] == vecData[3]) ||
			(vecData[0] == ATOM_TEST_DIFF && (*(wb_current.find(vecData[1])))[vecData[2]] != vecData[3])) ;
		else if((vecData[0] == VECTOR_TEST_IN && find(mySet.begin(), mySet.end(), vecData[3] ) != mySet.end()) ||
			(vecData[0] == VECTOR_TEST_NOT_IN && find(mySet.begin(), mySet.end(), vecData[3] ) == mySet.end())) ;
		else containState = false;
	}
	atEnd = containState;
	if(stateResult.empty()) {
		return false;
	}
	else {
		bool merged, constant;
		vector<TInterval> STResult;
		STResult.push_back(stateResult[0]); 
		stateResult.erase(stateResult.begin());
		vector<TInterval>::iterator it1, it2, it3;
		for(it1=stateResult.begin() ; it1!=stateResult.end() ; ++it1) {
			merged = false;
			for(it2=STResult.begin() ; it2!=STResult.end() ; ++it2)
				if(overlap(*it1,*it2)) {
					*it2 = hull(*it1,*it2);
					merged = true;
					break;
				}
			if(merged == false) STResult.push_back(*it1);
			do {
				constant = true;
				for(it2=STResult.begin() ; it2!=STResult.end() && constant ; ++it2)
					for(it3=it2+1 ; it3!=STResult.end() && constant ; ++it3)
						if(overlap(*it2, *it3)) {
							*it2 = hull(*it2,*it3);
							STResult.erase(it3);
							constant = false;
						}
			} while(!constant);
		}
		for(unsigned int i=0 ; i<STResult.size() ; i++) {
			durations.push_back(width(STResult[i]));
		}
	}
	return true;
}

vector<SocialRuleData*> TimeProjection::checkSubList(vector<Substitution*> const& vecSub, LightWorldBase const& wb, TaskList const& subTL) const {
	vector<SocialRuleData*> vecExit;
	unsigned int nb;
	TaskList* RebuiltSubTL;
	for(unsigned int i=0 ; i<vecSub.size() ; ++i) {
		RebuiltSubTL = subTL.Rebuild(*vecSub[i], wb);
		nb = projection.checkSubList(*RebuiltSubTL, projectionTimes);
		delete RebuiltSubTL; 
		if(nb > 0) vecExit.push_back(new SocialRuleData(vecSub[i], nb));
		else delete vecSub[i];
	}
	return vecExit;
}

TimeProjection::~TimeProjection() {}

