#if !defined HATP_HTNBASE_HH_
#define HATP_HTNBASE_HH_

#include<string>
#include<map>

#include "Task.hh"
#include "BeliefMethod.hh"

class HTNBase {

private:
	Task defaultTask;
	std::map<std::string, Task*> database;
	BeliefMethod* beliefMethod;

public:
	explicit HTNBase();
	HTNBase(HTNBase const& htnb);
	std::map<std::string, Task*> const& getDatabase() const;
	std::string toString() const;
	void add(Task* e);
	bool existingTask(std::string const& s) const;
	Task& find(std::string const& s);
	Task const& findConst(std::string const& s);
	Task const& find(std::string const& s) const;
	void setBeliefMethod(BeliefMethod* bm);
	BeliefMethod* getBeliefMethod() const;
	HTNBase& operator=(HTNBase const& htnb2copy);
	~HTNBase();
};

/**
 * constructor: empty (defautTask is automatically set to an empty Task)
 */
inline HTNBase::HTNBase() {}

/**
 * return database
 */
inline std::map<std::string, Task*> const& HTNBase::getDatabase() const {
	return database;
}

/**
 * return the Task with the name == s, if not found prints an error and
 * return defaultTask
 */
inline Task& HTNBase::find(std::string const& s) {
	std::map<std::string, Task*>::iterator it = database.find(s);
	if(it != database.end()) return *(it->second);
	std::cerr << "ERROR: trying to access to unknown task \"" << s << "\"" << std::endl;
	std::cerr << "default task return instead" << std::endl;
	return defaultTask;
}

/**
 * return the Task with the name == s, if not found prints an error and
 * return defaultTask (const version)
 */
inline Task const& HTNBase::find(std::string const& s) const {
	std::map<std::string, Task*>::const_iterator it = database.find(s);
	if(it != database.end()) return *(it->second);
	std::cerr << "ERROR: trying to access to unknown task \"" << s << "\"" << std::endl;
	std::cerr << "default task return instead" << std::endl;
	return defaultTask;
}

/**
 * return a pointer on a const version of Task with the name == s
 */
inline Task const& HTNBase::findConst(std::string const& s) {
	std::map<std::string, Task*>::const_iterator it = database.find(s);
	if(it != database.end()) return *(it->second);
	std::cerr << "ERROR: trying to access to unknown task \"" << s << "\"" << std::endl;
	std::cerr << "default task return instead" << std::endl;
	return defaultTask;
}

#endif
