#if !defined HATP_PLAN_SENDER_HH_
#define HATP_PLAN_SENDER_HH_

#include<queue>
#include<boost/thread/mutex.hpp>
#include<boost/utility.hpp>
#include<boost/variant/get.hpp>
#include<cassert>
#include<unistd.h>

#include "msgTypesTrait.hh"

template <typename MSG_TYPES, typename MSG_TYPES_TRAIT = MsgTypesTrait<MSG_TYPES> >
class PlanSender : boost::noncopyable {

public:

	explicit PlanSender() {
		if (pipe(mAtoBpipeFD) == -1) {
			perror("init");
			assert(false);
		}
		if (pipe(mBtoApipeFD) == -1) {
			perror("init");
			assert(false);
		}
	}

	~PlanSender() {
		close(mAtoBpipeFD[0]);
		close(mAtoBpipeFD[1]);
		close(mBtoApipeFD[0]);
		close(mBtoApipeFD[1]);
	}

	template <typename T> void sendObjectToB(const T &obj, MSG_TYPES msg) {
		boost::mutex::scoped_lock protect(mMutex);
		mAtoB.push(obj);
		sendMessageToB(msg);
	}

	template <typename T> void sendObjectToA(const T &obj, MSG_TYPES msg) {
		boost::mutex::scoped_lock protect(mMutex);
		mBtoA.push(obj);
		sendMessageToA(msg);
	}

	template <typename T> T receiveObjectFromB() {
		assert(objectAvailableFromB());
		typename MSG_TYPES_TRAIT::messagingToA_type res;
		boost::mutex::scoped_lock protect(mMutex);
		res = mBtoA.front();
		mBtoA.pop();
		return boost::get<T>(res);
	}

	template <typename T> T receiveObjectFromA() {
		assert(objectAvailableFromA());
		typename MSG_TYPES_TRAIT::messagingToB_type res;
		boost::mutex::scoped_lock protect(mMutex);
		res = mAtoB.front();
		mAtoB.pop();
		return boost::get<T>(res);
	}

	bool objectAvailableFromA() const {
		boost::mutex::scoped_lock protect(mMutex);
		return !mAtoB.empty();
	}

	bool objectAvailableFromB() const {
		boost::mutex::scoped_lock protect(mMutex);
		return !mBtoA.empty();
	}

	int fdToWaitA() const {
		return mAtoBpipeFD[0];
	}

	int fdToWaitB() const {
		return mBtoApipeFD[0];
	}

	void sendMessageToB(MSG_TYPES theMsg) const {
		sendMessage(mAtoBpipeFD[1], theMsg);
	}

	void sendMessageToA(MSG_TYPES theMsg) const {
		sendMessage(mBtoApipeFD[1], theMsg);
	}

	MSG_TYPES typeOfMessageFromA() const {
		return typeOfMessage(mAtoBpipeFD[0]);
	}

	MSG_TYPES typeOfMessageFromB() const {
		return typeOfMessage(mBtoApipeFD[0]);
	}

protected:

	MSG_TYPES typeOfMessage(int fd) const {
		char buffer[1];
		if (read(fd, buffer, 1) != 1) assert(false);
		return MSG_TYPES_TRAIT::fromChar(buffer[0]);
	}

	void sendMessage(int fd, MSG_TYPES m) const {
		char buffer[1];
		buffer[0] = MSG_TYPES_TRAIT::toChar(m);
		if (write(fd, buffer, 1) != 1) assert(false);
	}

private:

	std::queue<typename MSG_TYPES_TRAIT::messagingToB_type> mAtoB;
	std::queue<typename MSG_TYPES_TRAIT::messagingToA_type> mBtoA;
	mutable boost::mutex mMutex;
	int mAtoBpipeFD[2];
	int mBtoApipeFD[2];

};

#endif
